@@ -14,9 +14,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 #include <stdio.h>
 #include <stdlib.h>
@@ -30,14 +30,19 @@
 #endif
 #include <assert.h>
 #include <poll.h>
+#include <sys/statvfs.h>
+#include <sys/types.h>
 
 #include <libmount.h>
 
 #include "pathnames.h"
 #include "nls.h"
+#include "closestream.h"
 #include "c.h"
 #include "tt.h"
 #include "strutils.h"
+#include "xalloc.h"
+#include "optutils.h"
 
 /* flags */
 enum {
@@ -48,7 +53,9 @@ enum {
 	FL_NOSWAPMATCH	= (1 << 6),
 	FL_NOFSROOT	= (1 << 7),
 	FL_SUBMOUNTS	= (1 << 8),
-	FL_POLL		= (1 << 9)
+	FL_POLL		= (1 << 9),
+	FL_DF		= (1 << 10),
+	FL_ALL		= (1 << 11)
 };
 
 /* column IDs */
@@ -61,10 +68,18 @@ enum {
 	COL_FS_OPTIONS,
 	COL_LABEL,
 	COL_UUID,
+	COL_PARTLABEL,
+	COL_PARTUUID,
 	COL_MAJMIN,
 	COL_ACTION,
 	COL_OLD_TARGET,
 	COL_OLD_OPTIONS,
+	COL_SIZE,
+	COL_AVAIL,
+	COL_USED,
+	COL_USEPERC,
+	COL_FSROOT,
+	COL_TID,
 
 	FINDMNT_NCOLUMNS
 };
@@ -82,6 +97,7 @@ struct colinfo {
 	int		flags;		/* tt flags */
 	const char      *help;		/* column description */
 	const char	*match;		/* pattern for match_func() */
+	void		*match_data;	/* match specific data */
 };
 
 /* columns descriptions (don't use const, this is writable) */
@@ -94,10 +110,18 @@ static struct colinfo infos[FINDMNT_NCOLUMNS] = {
 	[COL_FS_OPTIONS]   = { "FS-OPTIONS",   0.10, TT_FL_TRUNC, N_("FS specific mount options") },
 	[COL_LABEL]        = { "LABEL",        0.10, 0, N_("filesystem label") },
 	[COL_UUID]         = { "UUID",           36, 0, N_("filesystem UUID") },
+	[COL_PARTLABEL]    = { "PARTLABEL",    0.10, 0, N_("partition label") },
+	[COL_PARTUUID]     = { "PARTUUID",       36, 0, N_("partition UUID") },
 	[COL_MAJMIN]       = { "MAJ:MIN",         6, 0, N_("major:minor device number") },
 	[COL_ACTION]       = { "ACTION",         10, TT_FL_STRICTWIDTH, N_("action detected by --poll") },
 	[COL_OLD_OPTIONS]  = { "OLD-OPTIONS",  0.10, TT_FL_TRUNC, N_("old mount options saved by --poll") },
 	[COL_OLD_TARGET]   = { "OLD-TARGET",   0.30, 0, N_("old mountpoint saved by --poll") },
+	[COL_SIZE]         = { "SIZE",            5, TT_FL_RIGHT, N_("filesystem size") },
+	[COL_AVAIL]        = { "AVAIL",           5, TT_FL_RIGHT, N_("filesystem size available") },
+	[COL_USED]         = { "USED",            5, TT_FL_RIGHT, N_("filesystem size used") },
+	[COL_USEPERC]      = { "USE%",            3, TT_FL_RIGHT, N_("filesystem use percentage") },
+	[COL_FSROOT]       = { "FSROOT",       0.25, TT_FL_NOEXTREMES, N_("filesystem root") },
+	[COL_TID]          = { "TID",             4, TT_FL_RIGHT, N_("task ID") },
 };
 
 /* global flags */
@@ -155,12 +179,64 @@ static const char *get_match(int id)
 	return infos[id].match;
 }
 
+static void *get_match_data(int id)
+{
+	assert(id < FINDMNT_NCOLUMNS);
+	return infos[id].match_data;
+}
+
 static void set_match(int id, const char *match)
 {
 	assert(id < FINDMNT_NCOLUMNS);
 	infos[id].match = match;
 }
 
+static void set_match_data(int id, void *data)
+{
+	assert(id < FINDMNT_NCOLUMNS);
+	infos[id].match_data = data;
+}
+
+/*
+ * source match means COL_SOURCE *or* COL_MAJMIN, depends on
+ * data format.
+ */
+static void set_source_match(const char *data)
+{
+	int maj, min;
+
+	if (sscanf(data, "%d:%d", &maj, &min) == 2) {
+		dev_t *devno = xmalloc(sizeof(dev_t));
+
+		*devno = makedev(maj, min);
+		set_match(COL_MAJMIN, data);
+		set_match_data(COL_MAJMIN, (void *) devno);
+		flags |= FL_NOSWAPMATCH;
+	} else
+		set_match(COL_SOURCE, data);
+}
+
+static void enable_extra_target_match(void)
+{
+	char *cn = NULL, *mnt = NULL;
+
+	/*
+	 * Check if match pattern is mountpoint, if not use the
+	 * real mountpoint.
+	 */
+	cn = mnt_resolve_path(get_match(COL_TARGET), cache);
+	if (!cn)
+		return;
+
+	mnt = mnt_get_mountpoint(cn);
+	if (!mnt || strcmp(mnt, cn) == 0)
+		return;
+
+	/* replace the current setting with the real mountpoint */
+	set_match(COL_TARGET, mnt);
+}
+
+
 static int is_tabdiff_column(int id)
 {
 	assert(id < FINDMNT_NCOLUMNS);
@@ -181,10 +257,14 @@ static int is_tabdiff_column(int id)
  */
 static int is_listall_mode(void)
 {
+	if ((flags & FL_DF) && !(flags & FL_ALL))
+		return 0;
+
 	return (!get_match(COL_SOURCE) &&
 		!get_match(COL_TARGET) &&
 		!get_match(COL_FSTYPE) &&
-		!get_match(COL_OPTIONS));
+		!get_match(COL_OPTIONS) &&
+		!get_match(COL_MAJMIN));
 }
 
 /*
@@ -280,14 +360,47 @@ static const char *get_tag(struct libmnt_fs *fs, const char *tagname)
 	return res;
 }
 
+static const char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
+{
+	struct statvfs buf;
+	uint64_t vfs_attr = 0;
+	char *sizestr;
+
+	if (statvfs(mnt_fs_get_target(fs), &buf) != 0)
+		return NULL;
+
+	switch(sizetype) {
+	case COL_SIZE:
+		vfs_attr = buf.f_frsize * buf.f_blocks;
+		break;
+	case COL_AVAIL:
+		vfs_attr = buf.f_frsize * buf.f_bavail;
+		break;
+	case COL_USED:
+		vfs_attr = buf.f_frsize * (buf.f_blocks - buf.f_bfree);
+		break;
+	case COL_USEPERC:
+		if (buf.f_blocks == 0)
+			return "-";
+
+		xasprintf(&sizestr, "%.0f%%",
+				(double)(buf.f_blocks - buf.f_bfree) /
+				buf.f_blocks * 100);
+		return sizestr;
+	}
+
+	return vfs_attr == 0 ? "0" :
+		size_to_human_string(SIZE_SUFFIX_1LETTER, vfs_attr);
+}
+
 /* reads FS data from libmount
- * TODO: add function that will deallocate data allocated by get_data()
  */
 static const char *get_data(struct libmnt_fs *fs, int num)
 {
 	const char *str = NULL;
+	int col_id = get_column_id(num);
 
-	switch(get_column_id(num)) {
+	switch (col_id) {
 	case COL_SOURCE:
 	{
 		const char *root = mnt_fs_get_root(fs);
@@ -305,7 +418,7 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 		if (root && str && !(flags & FL_NOFSROOT) && strcmp(root, "/")) {
 			char *tmp;
 
-			if (asprintf(&tmp, "%s[%s]", str, root) > 0)
+			if (xasprintf(&tmp, "%s[%s]", str, root) > 0)
 				str = tmp;
 		}
 		break;
@@ -328,9 +441,16 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 	case COL_UUID:
 		str = get_tag(fs, "UUID");
 		break;
+	case COL_PARTUUID:
+		str = get_tag(fs, "PARTUUID");
+		break;
 	case COL_LABEL:
 		str = get_tag(fs, "LABEL");
 		break;
+	case COL_PARTLABEL:
+		str = get_tag(fs, "PARTLABEL");
+		break;
+
 	case COL_MAJMIN:
 	{
 		dev_t devno = mnt_fs_get_devno(fs);
@@ -338,15 +458,32 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 			char *tmp;
 			int rc = 0;
 			if ((tt_flags & TT_FL_RAW) || (tt_flags & TT_FL_EXPORT))
-				rc = asprintf(&tmp, "%u:%u",
+				rc = xasprintf(&tmp, "%u:%u",
 					      major(devno), minor(devno));
 			else
-				rc = asprintf(&tmp, "%3u:%-3u",
+				rc = xasprintf(&tmp, "%3u:%-3u",
 					      major(devno), minor(devno));
 			if (rc)
 				str = tmp;
 		}
+		break;
 	}
+	case COL_SIZE:
+	case COL_AVAIL:
+	case COL_USED:
+	case COL_USEPERC:
+		str = get_vfs_attr(fs, col_id);
+		break;
+	case COL_FSROOT:
+		str = mnt_fs_get_root(fs);
+		break;
+	case COL_TID:
+		if (mnt_fs_get_tid(fs)) {
+			char *tmp;
+			if (xasprintf(&tmp, "%d", mnt_fs_get_tid(fs)) > 0)
+				str = tmp;
+		}
+		break;
 	default:
 		break;
 	}
@@ -496,17 +633,19 @@ static int parser_errcb(struct libmnt_table *tb __attribute__ ((__unused__)),
 
 static char **append_tabfile(char **files, int *nfiles, char *filename)
 {
-	void *tmp = realloc(files, sizeof(char *) * (*nfiles + 1));
-
-	if (!tmp) {
-		free(files);
-		return NULL;
-	}
-	files = tmp;
+	files = xrealloc(files, sizeof(char *) * (*nfiles + 1));
 	files[(*nfiles)++] = filename;
 	return files;
 }
 
+static char **append_pid_tabfile(char **files, int *nfiles, pid_t pid)
+{
+	char *path = NULL;
+
+	xasprintf(&path, "/proc/%d/mountinfo", (int) pid);
+	return append_tabfile(files, nfiles, path);
+}
+
 /* calls libmount fstab/mtab/mountinfo parser */
 static struct libmnt_table *parse_tabfiles(char **files,
 					   int nfiles,
@@ -570,12 +709,14 @@ static int tab_is_tree(struct libmnt_table *tb)
 	return rc;
 }
 
+
 /* filter function for libmount (mnt_table_find_next_fs()) */
 static int match_func(struct libmnt_fs *fs,
 		      void *data __attribute__ ((__unused__)))
 {
 	int rc = flags & FL_INVERT ? 1 : 0;
 	const char *m;
+	void *md;
 
 	m = get_match(COL_TARGET);
 	if (m && !mnt_fs_match_target(fs, m, cache))
@@ -593,6 +734,20 @@ static int match_func(struct libmnt_fs *fs,
 	if (m && !mnt_fs_match_options(fs, m))
 		return rc;
 
+	md = get_match_data(COL_MAJMIN);
+	if (md && mnt_fs_get_devno(fs) != *((dev_t *) md))
+		return rc;
+
+	if ((flags & FL_DF) && !(flags & FL_ALL)) {
+		const char *type = mnt_fs_get_fstype(fs);
+
+		if (type && strstr(type, "tmpfs"))	/* tmpfs is wanted */
+			return !rc;
+
+		if (mnt_fs_is_pseudofs(fs))
+			return rc;
+	}
+
 	return !rc;
 }
 
@@ -688,7 +843,7 @@ static int poll_match(struct libmnt_fs *fs)
 	    get_match(COL_SOURCE) && !get_match(COL_TARGET)) {
 		/*
 		 * findmnt --poll /foo
-		 * The '/foo' source as well as target.
+		 * The '/foo' maybe source as well as target.
 		 */
 		const char *str = get_match(COL_SOURCE);
 
@@ -738,7 +893,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 
 	f = fopen(tabfile, "r");
 	if (!f) {
-		warn(_("%s: open failed"), tabfile);
+		warn(_("cannot open %s"), tabfile);
 		goto done;
 	}
 
@@ -816,8 +971,8 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 {
 	int i;
 
+	fputs(USAGE_HEADER, out);
 	fprintf(out, _(
-	"\nUsage:\n"
 	" %1$s [options]\n"
 	" %1$s [options] <device> | <mountpoint>\n"
 	" %1$s [options] <device> <mountpoint>\n"
@@ -836,17 +991,20 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	" -w, --timeout <num>    upper limit in milliseconds that --poll will block\n\n"));
 
 	fprintf(out, _(
+	" -A, --all              disable all built-in filters, print all filesystems\n"
 	" -a, --ascii            use ASCII chars for tree formatting\n"
 	" -c, --canonicalize     canonicalize printed paths\n"
+	" -D, --df               imitate the output of df(1)\n"
 	" -d, --direction <word> direction of search, 'forward' or 'backward'\n"
-	" -e, --evaluate         convert tags (LABEL/UUID) to device names\n"
+	" -e, --evaluate         convert tags (LABEL,UUID,PARTUUID,PARTLABEL) \n"
+	"                          to device names\n"
 	" -F, --tab-file <path>  alternative file for --fstab, --mtab or --kernel options\n"
 	" -f, --first-only       print the first found filesystem only\n"));
 
 	fprintf(out, _(
-	" -h, --help             display this help text and exit\n"
 	" -i, --invert           invert the sense of matching\n"
 	" -l, --list             use list format output\n"
+	" -N, --task <tid>       use alternative namespace (/proc/<tid>/mountinfo file)\n"
 	" -n, --noheadings       don't print column headings\n"
 	" -u, --notruncate       don't truncate text in columns\n"));
 	fprintf(out, _(
@@ -858,9 +1016,13 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	fprintf(out, _(
 	" -v, --nofsroot         don't print [/dir] for bind or btrfs mounts\n"
 	" -R, --submounts        print all submounts for the matching filesystems\n"
-	" -S, --source <string>  the device to mount (by name, LABEL= or UUID=)\n"
-	" -T, --target <string>  the mountpoint to use\n\n"));
+	" -S, --source <string>  the device to mount (by name, maj:min, \n"
+	"                          LABEL=, UUID=, PARTUUID=, PARTLABEL=)\n"
+	" -T, --target <string>  the mountpoint to use\n"));
 
+	fputs(USAGE_SEPARATOR, out);
+	fputs(USAGE_HELP, out);
+	fputs(USAGE_VERSION, out);
 
 	fprintf(out, _("\nAvailable columns:\n"));
 
@@ -872,12 +1034,6 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
-static void __attribute__((__noreturn__))
-errx_mutually_exclusive(const char *opts)
-{
-	errx(EXIT_FAILURE, "%s %s", opts, _("options are mutually exclusive"));
-}
-
 int main(int argc, char *argv[])
 {
 	struct libmnt_table *tb = NULL;
@@ -885,14 +1041,16 @@ int main(int argc, char *argv[])
 	int direction = MNT_ITER_FORWARD;
 	int i, c, rc = -1, timeout = -1;
 	int ntabfiles = 0, tabtype = 0;
+	char *outarg = NULL;
 
-	/* table.h */
 	struct tt *tt = NULL;
 
 	static const struct option longopts[] = {
+	    { "all",          0, 0, 'A' },
 	    { "ascii",        0, 0, 'a' },
 	    { "canonicalize", 0, 0, 'c' },
 	    { "direction",    1, 0, 'd' },
+	    { "df",           0, 0, 'D' },
 	    { "evaluate",     0, 0, 'e' },
 	    { "first-only",   0, 0, 'f' },
 	    { "fstab",        0, 0, 's' },
@@ -913,31 +1071,52 @@ int main(int argc, char *argv[])
 	    { "submounts",    0, 0, 'R' },
 	    { "source",       1, 0, 'S' },
 	    { "tab-file",     1, 0, 'F' },
+	    { "task",         1, 0, 'N' },
 	    { "target",       1, 0, 'T' },
 	    { "timeout",      1, 0, 'w' },
+	    { "version",      0, 0, 'V' },
 
 	    { NULL,           0, 0, 0 }
 	};
 
+	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
+		{ 'N','k','m','s' },		/* task,kernel,mtab,fstab */
+		{ 'P','l','r' },		/* pairs,list,raw */
+		{ 'm','p','s' },		/* mtab,poll,fstab */
+		{ 0 }
+	};
+	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
+
 	assert(ARRAY_SIZE(columns) == FINDMNT_NCOLUMNS);
 
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	/* default output format */
 	tt_flags |= TT_FL_TREE;
 
 	while ((c = getopt_long(argc, argv,
-				"acd:ehifF:o:O:p::Pklmnrst:uvRS:T:w:",
+				"AacDd:ehifF:o:O:p::PklmnN:rst:uvRS:T:w:V",
 				longopts, NULL)) != -1) {
+
+		err_exclusive_options(c, longopts, excl, excl_st);
+
 		switch(c) {
+		case 'A':
+			flags |= FL_ALL;
+			break;
 		case 'a':
 			tt_flags |= TT_FL_ASCII;
 			break;
 		case 'c':
 			flags |= FL_CANONICALIZE;
 			break;
+		case 'D':
+			tt_flags &= ~TT_FL_TREE;
+			flags |= FL_DF;
+			break;
 		case 'd':
 			if (!strcmp(optarg, "forward"))
 				direction = MNT_ITER_FORWARD;
@@ -966,11 +1145,7 @@ int main(int argc, char *argv[])
 			disable_columns_truncate();
 			break;
 		case 'o':
-			ncolumns = string_to_idarray(optarg,
-						columns, ARRAY_SIZE(columns),
-						column_name_to_id);
-			if (ncolumns < 0)
-				exit(EXIT_FAILURE);
+			outarg = optarg;
 			break;
 		case 'O':
 			set_match(COL_OPTIONS, optarg);
@@ -991,20 +1166,14 @@ int main(int argc, char *argv[])
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 'm':		/* mtab */
-			if (tabtype)
-				errx_mutually_exclusive("--{fstab,mtab,kernel}");
 			tabtype = TABTYPE_MTAB;
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 's':		/* fstab */
-			if (tabtype)
-				errx_mutually_exclusive("--{fstab,mtab,kernel}");
 			tabtype = TABTYPE_FSTAB;
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 'k':		/* kernel (mountinfo) */
-			if (tabtype)
-				 errx_mutually_exclusive("--{fstab,mtab,kernel}");
 			tabtype = TABTYPE_KERNEL;
 			break;
 		case 't':
@@ -1015,14 +1184,17 @@ int main(int argc, char *argv[])
 			tt_flags |= TT_FL_RAW;		/* enable raw */
 			break;
 		case 'l':
-			if ((tt_flags & TT_FL_RAW) && (tt_flags & TT_FL_EXPORT))
-				errx_mutually_exclusive("--{raw,list,pairs}");
-
 			tt_flags &= ~TT_FL_TREE; /* disable the default */
 			break;
 		case 'n':
 			tt_flags |= TT_FL_NOHEADINGS;
 			break;
+		case 'N':
+			tabtype = TABTYPE_KERNEL;
+			tabfiles = append_pid_tabfile(tabfiles, &ntabfiles,
+					strtou32_or_err(optarg,
+						_("invalid TID argument")));
+			break;
 		case 'v':
 			flags |= FL_NOFSROOT;
 			break;
@@ -1030,7 +1202,7 @@ int main(int argc, char *argv[])
 			flags |= FL_SUBMOUNTS;
 			break;
 		case 'S':
-			set_match(COL_SOURCE, optarg);
+			set_source_match(optarg);
 			flags |= FL_NOSWAPMATCH;
 			break;
 		case 'T':
@@ -1038,15 +1210,27 @@ int main(int argc, char *argv[])
 			flags |= FL_NOSWAPMATCH;
 			break;
 		case 'w':
-			timeout = strtol_or_err(optarg,
-					_("failed to parse timeout"));
+			timeout = strtos32_or_err(optarg, _("invalid timeout argument"));
 			break;
+		case 'V':
+			printf(UTIL_LINUX_VERSION);
+			return EXIT_SUCCESS;
 		default:
 			usage(stderr);
 			break;
 		}
 	}
 
+	if (!ncolumns && (flags & FL_DF)) {
+		columns[ncolumns++] = COL_SOURCE;
+		columns[ncolumns++] = COL_FSTYPE;
+		columns[ncolumns++] = COL_SIZE;
+		columns[ncolumns++] = COL_USED;
+		columns[ncolumns++] = COL_AVAIL;
+		columns[ncolumns++] = COL_USEPERC;
+		columns[ncolumns++] = COL_TARGET;
+	}
+
 	/* default columns */
 	if (!ncolumns) {
 		if (flags & FL_POLL)
@@ -1058,16 +1242,15 @@ int main(int argc, char *argv[])
 		columns[ncolumns++] = COL_OPTIONS;
 	}
 
+	if (outarg && string_add_to_idarray(outarg, columns, ARRAY_SIZE(columns),
+					 &ncolumns, column_name_to_id) < 0)
+		return EXIT_FAILURE;
+
 	if (!tabtype)
 		tabtype = TABTYPE_KERNEL;
 
-
-	if (flags & FL_POLL) {
-		if (tabtype != TABTYPE_KERNEL)
-			errx_mutually_exclusive("--{poll,fstab,mtab}");
-		if (ntabfiles > 1)
-			errx(EXIT_FAILURE, _("--poll accepts only one file, but mure specified by --tab-file"));
-	}
+	if ((flags & FL_POLL) && ntabfiles > 1)
+		errx(EXIT_FAILURE, _("--poll accepts only one file, but more specified by --tab-file"));
 
 	if (optind < argc && (get_match(COL_SOURCE) || get_match(COL_TARGET)))
 		errx(EXIT_FAILURE, _(
@@ -1075,7 +1258,7 @@ int main(int argc, char *argv[])
 			"with command line element that is not an option"));
 
 	if (optind < argc)
-		set_match(COL_SOURCE, argv[optind++]);	/* dev/tag/mountpoint */
+		set_source_match(argv[optind++]);	/* dev/tag/mountpoint/maj:min */
 	if (optind < argc)
 		set_match(COL_TARGET, argv[optind++]);	/* mountpoint */
 
@@ -1095,7 +1278,8 @@ int main(int argc, char *argv[])
 		 */
 		const char *x = get_match(COL_SOURCE);
 
-		if (!strncmp(x, "LABEL=", 6) || !strncmp(x, "UUID=", 5))
+		if (!strncmp(x, "LABEL=", 6) || !strncmp(x, "UUID=", 5) ||
+		    !strncmp(x, "PARTLABEL=", 10) || !strncmp(x, "PARTUUID=", 9))
 			flags |= FL_NOSWAPMATCH;
 	}
 
@@ -1108,7 +1292,7 @@ int main(int argc, char *argv[])
 	if (!tb)
 		goto leave;
 
-	if ((tt_flags & TT_FL_TREE) && !tab_is_tree(tb))
+	if ((tt_flags & TT_FL_TREE) && (ntabfiles > 1 || !tab_is_tree(tb)))
 		tt_flags &= ~TT_FL_TREE;
 
 	cache = mnt_new_cache();
@@ -1118,6 +1302,14 @@ int main(int argc, char *argv[])
 	}
 	mnt_table_set_cache(tb, cache);
 
+	if (tabtype == TABTYPE_KERNEL
+	    && (flags & FL_NOSWAPMATCH)
+	    && get_match(COL_TARGET))
+		/*
+		 * enable extra functionality for target match
+		 */
+		enable_extra_target_match();
+
 	/*
 	 * initialize output formatting (tt.h)
 	 */
