@@ -45,6 +45,7 @@
 #include "usleep.h"
 #include "strutils.h"
 #include "c.h"
+#include "closestream.h"
 
 #define DEFAULT_LINES  10
 
@@ -58,7 +59,7 @@ tailf(const char *filename, int lines)
 	int  i;
 
 	if (!(str = fopen(filename, "r")))
-		err(EXIT_FAILURE, _("cannot open \"%s\" for read"), filename);
+		err(EXIT_FAILURE, _("cannot open %s"), filename);
 
 	buf = xmalloc((lines ? lines : 1) * BUFSIZ);
 	p = buf;
@@ -95,10 +96,10 @@ roll_file(const char *filename, off_t *size)
 
 	fd = open(filename, O_RDONLY);
 	if (fd < 0)
-		err(EXIT_FAILURE, _("cannot open \"%s\" for read"), filename);
+		err(EXIT_FAILURE, _("cannot open %s"), filename);
 
 	if (fstat(fd, &st) == -1)
-		err(EXIT_FAILURE, _("cannot stat \"%s\""), filename);
+		err(EXIT_FAILURE, _("stat failed %s"), filename);
 
 	if (st.st_size == *size) {
 		close(fd);
@@ -209,7 +210,7 @@ static void __attribute__ ((__noreturn__)) usage(FILE *out)
 }
 
 /* parses -N option */
-long old_style_option(int *argc, char **argv)
+static long old_style_option(int *argc, char **argv)
 {
 	int i = 1, nargs = *argc;
 	long lines = -1;
@@ -240,6 +241,7 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	static const struct option longopts[] = {
 		{ "lines",   required_argument, 0, 'n' },
@@ -275,7 +277,7 @@ int main(int argc, char **argv)
 	filename = argv[optind];
 
 	if (stat(filename, &st) != 0)
-		err(EXIT_FAILURE, _("cannot stat \"%s\""), filename);
+		err(EXIT_FAILURE, _("stat failed %s"), filename);
 
 	size = st.st_size;;
 	tailf(filename, lines);
